@@ -41,6 +41,7 @@ struct entry_data {
 	char *domain;
 	char *server;
 	unsigned int flags;
+	guint timeout;
 };
 
 static GSList *entry_list = NULL;
@@ -60,6 +61,8 @@ static void remove_entries(GSList *entries)
 			resolver->remove(entry->interface, entry->domain,
 								entry->server);
 
+		if (entry->timeout)
+			g_source_remove(entry->timeout);
 		g_free(entry->server);
 		g_free(entry->domain);
 		g_free(entry->interface);
@@ -137,16 +140,31 @@ void connman_resolver_unregister(struct connman_resolver *resolver)
 	remove_entries(matches);
 }
 
+static gboolean resolver_expire_cb(gpointer user_data)
+{
+	struct entry_data *entry = user_data;
+	GSList *list;
+
+	DBG("interface %s domain %s server %s",
+	    entry->interface, entry->domain, entry->server);
+
+	list = g_slist_append(NULL, entry);
+	remove_entries(list);
+
+	return FALSE;
+}
+
 static int append_resolver(const char *interface, const char *domain,
-					const char *server, unsigned int flags)
+			   const char *server, unsigned int lifetime,
+			   unsigned int flags)
 {
 	struct entry_data *entry;
 	GSList *list;
 
-	DBG("interface %s domain %s server %s flags %d",
-					interface, domain, server, flags);
+	DBG("interface %s domain %s server %s lifetime %d flags %d",
+	    interface, domain, server, lifetime, flags);
 
-	if (server == NULL)
+	if (server == NULL && domain == NULL)
 		return -EINVAL;
 
 	entry = g_try_new0(struct entry_data, 1);
@@ -157,6 +175,10 @@ static int append_resolver(const char *interface, const char *domain,
 	entry->domain = g_strdup(domain);
 	entry->server = g_strdup(server);
 	entry->flags = flags;
+	if (lifetime)
+		entry->timeout = g_timeout_add_seconds(lifetime,
+						       resolver_expire_cb,
+						       entry);
 
 	entry_list = g_slist_append(entry_list, entry);
 
@@ -188,7 +210,45 @@ int connman_resolver_append(const char *interface, const char *domain,
 {
 	DBG("interface %s domain %s server %s", interface, domain, server);
 
-	return append_resolver(interface, domain, server, 0);
+	return append_resolver(interface, domain, server, 0, 0);
+}
+
+/**
+ * connman_resolver_append_lifetime:
+ * @interface: network interface
+ * @domain: domain limitation
+ * @server: server address
+ * @timeout: server lifetime in seconds
+ *
+ * Append resolver server address to current list
+ */
+int connman_resolver_append_lifetime(const char *interface, const char *domain,
+				     const char *server, unsigned int lifetime)
+{
+	GSList *list;
+
+	DBG("interface %s domain %s server %s lifetime %d",
+	    interface, domain, server, lifetime);
+
+	if (server == NULL)
+		return -EINVAL;
+
+	for (list = entry_list; list; list = list->next) {
+		struct entry_data *entry = list->data;
+
+		if (!entry->timeout ||
+		    g_strcmp0(entry->interface, interface) ||
+		    g_strcmp0(entry->domain, domain) ||
+		    g_strcmp0(entry->server, server))
+			continue;
+
+		g_source_remove(entry->timeout);
+		entry->timeout = g_timeout_add_seconds(lifetime,
+						       resolver_expire_cb,
+						       entry);
+		return 0;
+	}
+	return append_resolver(interface, domain, server, lifetime, 0);
 }
 
 /**
@@ -275,7 +335,7 @@ int connman_resolver_append_public_server(const char *server)
 {
 	DBG("server %s", server);
 
-	return append_resolver(NULL, NULL, server, RESOLVER_FLAG_PUBLIC);
+	return append_resolver(NULL, NULL, server, 0, RESOLVER_FLAG_PUBLIC);
 }
 
 /**
@@ -312,49 +372,6 @@ void connman_resolver_flush(void)
 	return;
 }
 
-static int selftest_append(const char *interface, const char *domain,
-							const char *server)
-{
-	DBG("server %s", server);
-
-	return 0;
-}
-
-static int selftest_remove(const char *interface, const char *domain,
-							const char *server)
-{
-	DBG("server %s", server);
-
-	return 0;
-}
-
-static struct connman_resolver selftest_resolver = {
-	.name     = "selftest",
-	.priority = CONNMAN_RESOLVER_PRIORITY_HIGH + 42,
-	.append   = selftest_append,
-	.remove   = selftest_remove,
-};
-
-int __connman_resolver_selftest(void)
-{
-	connman_resolver_append("wlan0", "lwn.net", "192.168.0.1");
-
-	connman_resolver_register(&selftest_resolver);
-
-	connman_resolver_append("eth0", "connman.net", "192.168.42.1");
-	connman_resolver_append("wlan0", "lwn.net", "192.168.0.2");
-
-	connman_resolver_append_public_server("8.8.8.8");
-
-	connman_resolver_remove_public_server("8.8.8.8");
-
-	connman_resolver_remove_all("wlan0");
-
-	connman_resolver_unregister(&selftest_resolver);
-
-	return 0;
-}
-
 struct resolvfile_entry {
 	char *interface;
 	char *domain;
@@ -392,14 +409,37 @@ static int resolvfile_export(void)
 	content = g_string_new("# Generated by Connection Manager\n");
 
 	/*
-	 * Nameservers are added in reverse so that the most recently
-	 * appended entry is the primary nameserver. No more than MAXNS
-	 * nameservers are used.
+	 * Domains and nameservers are added in reverse so that the most
+	 * recently appended entry is the primary one. No more than
+	 * MAXDNSRCH/MAXNS entries are used.
 	 */
+
+	for (count = 0, list = g_list_last(resolvfile_list);
+						list && (count < MAXDNSRCH);
+						list = g_list_previous(list)) {
+		struct resolvfile_entry *entry = list->data;
+
+		if (!entry->domain)
+			continue;
+
+		if (count == 0)
+			g_string_append_printf(content, "search ");
+
+		g_string_append_printf(content, "%s ", entry->domain);
+		count++;
+	}
+
+	if (count)
+		g_string_append_printf(content, "\n");
+
 	for (count = 0, list = g_list_last(resolvfile_list);
 						list && (count < MAXNS);
 						list = g_list_previous(list)) {
 		struct resolvfile_entry *entry = list->data;
+
+		if (!entry->server)
+			continue;
+
 		g_string_append_printf(content, "nameserver %s\n",
 								entry->server);
 		count++;
