@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -39,6 +39,7 @@ struct notify_data {
 struct connman_task {
 	char *path;
 	pid_t pid;
+	guint child_watch;
 	GPtrArray *argv;
 	GPtrArray *envp;
 	connman_task_exit_t exit_func;
@@ -69,6 +70,9 @@ static void free_task(gpointer data)
 	if (task->pid > 0)
 		kill(task->pid, SIGTERM);
 
+	if (task->child_watch > 0)
+		g_source_remove(task->child_watch);
+
 	g_ptr_array_foreach(task->envp, free_pointer, NULL);
 	g_ptr_array_free(task->envp, TRUE);
 
@@ -251,6 +255,8 @@ static void task_died(GPid pid, gint status, gpointer user_data)
 	g_spawn_close_pid(pid);
 	task->pid = -1;
 
+	task->child_watch = 0;
+
 	if (task->exit_func)
 		task->exit_func(task, task->exit_data);
 }
@@ -267,14 +273,16 @@ static void task_setup(gpointer user_data)
  * @task: task structure
  * @function: exit callback
  * @user_data: optional exit user data
+ * @fd: optional spawn with pipe
  *
  * Execute program specified by #task
  */
 int connman_task_run(struct connman_task *task,
-			connman_task_exit_t function, void *user_data)
+			connman_task_exit_t function, void *user_data,
+			int *stdin_fd, int *stdout_fd, int *stderr_fd)
 {
-	GSpawnFlags flags = G_SPAWN_DO_NOT_REAP_CHILD |
-						G_SPAWN_STDOUT_TO_DEV_NULL;
+	GSpawnFlags flags = G_SPAWN_DO_NOT_REAP_CHILD;
+	gboolean result;
 	char **argv, **envp;
 
 	DBG("task %p", task);
@@ -282,6 +290,12 @@ int connman_task_run(struct connman_task *task,
 	if (task->pid > 0)
 		return -EALREADY;
 
+	if (stdout_fd == NULL)
+		flags |= G_SPAWN_STDOUT_TO_DEV_NULL;
+
+	if (stderr_fd == NULL)
+		flags |= G_SPAWN_STDERR_TO_DEV_NULL;
+
 	task->exit_func = function;
 	task->exit_data = user_data;
 
@@ -312,13 +326,15 @@ int connman_task_run(struct connman_task *task,
 	argv = (char **) task->argv->pdata;
 	envp = (char **) task->envp->pdata;
 
-	if (g_spawn_async(NULL, argv, envp, flags,
-				task_setup, task, &task->pid, NULL) == FALSE) {
+	result = g_spawn_async_with_pipes(NULL, argv, envp, flags,
+					task_setup, task, &task->pid,
+					stdin_fd, stdout_fd, stderr_fd, NULL);
+	if (result == FALSE) {
 		connman_error("Failed to spawn %s", argv[0]);
 		return -EIO;
 	}
 
-	g_child_watch_add(task->pid, task_died, task);
+	task->child_watch = g_child_watch_add(task->pid, task_died, task);
 
 	return 0;
 }
